import re
from rdkit import Chem  # type: ignore


class ChemDoodle:
    """
    Utils to use [ChemDoolde JSON format]\
(https://web.chemdoodle.com/docs/chemdoodle-json-format)
    """

    @classmethod
    def mol_rdkit_to_json(cls, mr, begin_id={"a": 0, "b": 0}):

        json_res = {"a": [], "b": []}
        Chem.rdDepictor.Compute2DCoords(mr)
        ZOOM = 20
        positions = mr.GetConformer().GetPositions()
        BOND_TYPE_DIC = {
            Chem.rdchem.BondType.SINGLE: 1,
            Chem.rdchem.BondType.DOUBLE: 2,
            Chem.rdchem.BondType.TRIPLE: 3,
        }
        chiral_bonds = {}
        CHIRAL_CONFIG = {
            Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW: {
                0: "protruding",
                2: "recessed",
            },
            Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW: {0: "recessed", 2: "protruding"},
        }

        for i, a in enumerate(mr.GetAtoms()):
            mol_json = {
                "i": "a{0}".format(i + begin_id["a"]),
                "x": ZOOM * positions[i][0],
                "y": ZOOM * positions[i][1],
            }
            symbols = re.findall(r"\[([^:]*?)(?:\:\d)?\]", a.GetSmarts())
            if len(symbols) > 0:
                symbols = symbols[0].split(",")
            if len(symbols) > 1:
                v = []
                if "*" in symbols:
                    v = ["a"]
                else:
                    for sy in symbols:
                        find_at = re.findall(r"#(\d)", sy)
                        if len(find_at) == 1:
                            try:
                                at_value = int(find_at[0])
                            except Exception:
                                at_value = find_at[0]
                            v.append(Chem.Atom(at_value).GetSymbol())
                        else:
                            v.append(sy)
                mol_json["q"] = {"as": {"v": v, "n": False}}
            else:
                symbol = a.GetSymbol()
                if symbol != "C":
                    mol_json["l"] = symbol
                charge = a.GetFormalCharge()
                if charge != 0:
                    mol_json["c"] = charge
                elif a.GetChiralTag() != Chem.rdchem.ChiralType.CHI_UNSPECIFIED:
                    ct = a.GetChiralTag()
                    for i, b in enumerate(a.GetBonds()):
                        if i in CHIRAL_CONFIG[ct]:
                            chiral_bonds[b.GetIdx()] = CHIRAL_CONFIG[ct][i]
            json_res["a"].append(mol_json)

        try:
            # Force atom to be aromatic if is in ring
            for a in mr.GetAtoms():
                a.SetIsAromatic(a.IsInRing())
            # Kekulize mol
            Chem.rdmolops.SanitizeMol(mr)
            Chem.Kekulize(mr, True)

        except Exception:
            # Transform aromatic bonds in kekulize form
            # This should not be used ...
            arom_bond_type = Chem.rdchem.BondType.SINGLE

            def propagate_kekulize(atom, arom_bond_type):
                for b in atom.GetBonds():
                    if b.GetBondType() == Chem.rdchem.BondType.AROMATIC:
                        if arom_bond_type == Chem.rdchem.BondType.SINGLE:
                            arom_bond_type = Chem.rdchem.BondType.DOUBLE
                        else:
                            arom_bond_type = Chem.rdchem.BondType.SINGLE
                        b.SetBondType(arom_bond_type)
                        if atom.GetIdx() != b.GetBeginAtomIdx():
                            target = b.GetBeginAtom()
                        else:
                            target = b.GetEndAtom()
                        propagate_kekulize(target, arom_bond_type)

            for b in mr.GetBonds():
                if b.GetBondType() == Chem.rdchem.BondType.AROMATIC:
                    b.SetBondType(arom_bond_type)
                    propagate_kekulize(b.GetBeginAtom(), arom_bond_type)

        for i, b in enumerate(mr.GetBonds()):
            b_id = b.GetIdx()
            bond_json = {
                "i": "b{0}".format(i + begin_id["b"]),
                "b": b.GetBeginAtomIdx(),
                "e": b.GetEndAtomIdx(),
            }
            bond_type = b.GetBondType()
            if bond_type != Chem.rdchem.BondType.SINGLE:
                bond_json["o"] = BOND_TYPE_DIC[bond_type]
            if b_id in chiral_bonds:
                bond_json["s"] = chiral_bonds[b_id]
            json_res["b"].append(bond_json)
        return json_res

    @classmethod
    def react_to_json(cls, reaction):
        PADDING = 80
        ARROW_LENGTH = 60
        json_res = {
            "m": [],
            "s": [],
        }
        x_bound = 0
        atom_maping = {}
        begin_id = {"a": 0, "b": 0}

        def append_mol(m, mol_type):
            m_json = cls.mol_rdkit_to_json(m, begin_id)
            x_max = 0
            x_min = 0
            for a in m_json["a"]:
                x_min = min(x_min, a["x"])
            for a in m_json["a"]:
                a["x"] += x_bound - x_min
                x_max = max(x_max, a["x"])
            for i, a in enumerate(m.GetAtoms()):
                map_num = a.GetAtomMapNum()
                if map_num > 0:
                    atom_id = "a{0}".format(i + begin_id["a"])
                    if mol_type == "reactant":
                        atom_maping[map_num] = [atom_id, None]
                    elif mol_type == "product":
                        atom_maping[map_num][1] = atom_id
            json_res["m"].append(m_json)
            begin_id["a"] += len(m_json["a"])
            begin_id["b"] += len(m_json["b"])

            return x_max

        # Reactants
        for m in reaction.GetReactants():
            x_bound += append_mol(m, "reactant") + PADDING

        # Arrow
        x_bound -= PADDING / 2
        json_res["s"].append(
            {
                "i": "s0",
                "t": "Line",
                "x1": x_bound,
                "y1": 0,
                "x2": x_bound + ARROW_LENGTH,
                "y2": 0,
                "a": "synthetic",
            }
        )
        x_bound += ARROW_LENGTH + PADDING

        # Products
        for m in reaction.GetProducts():
            x_bound += append_mol(m, "product")

        # AtomMapping
        for i, map_ in atom_maping.items():
            json_res["s"].append(
                {
                    "i": "s{0}".format(i),
                    "t": "AtomMapping",
                    "a1": map_[0],
                    "a2": map_[1],
                }
            )

        return json_res



from __future__ import annotations
from typing import List, Tuple, Set
import re
import numpy as np
from rdkit import Chem
from rdkit.Chem import Descriptors
from rdkit.Chem.Draw import IPythonConsole

AROMATICITY_MODEL = Chem.rdmolops.AromaticityModel.AROMATICITY_MDL


class Molecule:
    """
    Instance of molecule using RDKit `Mol` class.

    It's instantiate by providing SMILES but it's
    also possible to use `from_rdkit()` class method
    to instantiate from RDKit `Mol` instance.

    !!! example
        ```python
        from chem_kit import Molecule
        mol = Molecule("CCO")
        ```
    """

    def __init__(
        self,
        smiles: str,
        aromaticity: str = "mdl",
        preserve_H: bool = False,
    ):
        """
        Args:
            smiles: SMILES of the molecule to create.
            aromaticity:
                Aromacity model to apply.

                Possible values are :

                - custom
                - default
                - mdl
                - rdkit
                - simple

                If None, default is "mdl".
            preserve_H: Preserve H that are explicit in smiles.

        """
        self._set_aromaticity_model(aromaticity)
        self._set_rdkit(smiles, preserve_H=preserve_H)
        self._smiles = Chem.MolToSmiles(self.rdkit)

    @property
    def smiles(self) -> str:
        """Cannonical SMILES of the molecule"""
        return self._smiles

    @property
    def rdkit(self) -> Chem.Mol:
        """
        `rdkit.Chem.Mol` instance of the molecule.
        Use it to directly apply RDKit methods
        """
        return self._rdkit.__copy__()

    @property
    def smiles_kekulized(self):
        """Kekulized SMILES of the molecule"""
        return Chem.MolToSmiles(self._kekulize_rdkit())

    @property
    def mass(self):
        return Descriptors.MolWt(self.rdkit)

    def _kekulize_rdkit(self):
        mol = self.rdkit
        Chem.Kekulize(mol, clearAromaticFlags=True)
        return mol

    def __eq__(self, other):
        """
        Comparaison betwwen two Molecule instances is done
        by comparing cannonical SMILES (`.smiles` property)
        """
        if not isinstance(other, Molecule):
            return NotImplemented

        return self.smiles == other.smiles

    @classmethod
    def from_rdkit(cls, rdkit: Chem.Mol) -> Molecule:
        """
        Instanciate by providing a RDKit instance rather than a smiles.

        Args:
            rdkit: `rdkit.Chem.Mol` instance

        Returns:
            Molecule instance
        """
        rdkit = cls.resolve_charge(rdkit)
        smiles = Chem.MolToSmiles(rdkit)
        return cls(smiles)

    @staticmethod
    def resolve_charge(rdkit):
        map_charge = re.findall(r"\[#\d+\&(\++):(\d+)]", Chem.MolToSmarts(rdkit))
        map_charge = {int(map_num): len(charge) for charge, map_num in map_charge}
        for atom in rdkit.GetAtoms():
            charge = map_charge.get(atom.GetAtomMapNum())
            if charge:
                atom.SetFormalCharge(charge)
        return rdkit

    def _set_aromaticity_model(self, aromaticity):
        self._AROMATICITY_MODEL = getattr(
            Chem.rdmolops.AromaticityModel,
            "AROMATICITY_{}".format(aromaticity.upper()),
        )

    def _set_rdkit(self, smiles, preserve_H):
        if preserve_H:
            mol = self._mol_from_smiles_preserve_H(smiles)
        else:
            mol = Chem.MolFromSmiles(smiles)
        Chem.Kekulize(mol, True)
        Chem.SetAromaticity(mol, self._AROMATICITY_MODEL)
        self._rdkit = mol

    @staticmethod
    def _mol_from_smiles_preserve_H(smiles):
        mol = Chem.MolFromSmiles(smiles, sanitize=False)
        sanitize_params = [
            # "ADJUSTHS",
            # "ALL" ,
            "CLEANUP",
            "CLEANUPCHIRALITY",
            "FINDRADICALS",
            "KEKULIZE",
            # "NONE",
            "PROPERTIES",
            "SETAROMATICITY",
            "SETCONJUGATION",
            "SETHYBRIDIZATION",
            "SYMMRINGS",
        ]
        arr = np.array(
            [
                getattr(Chem.rdmolops.SanitizeFlags, "SANITIZE_" + param)
                for param in sanitize_params
            ]
        )
        sanitize_param = int(np.bitwise_or.reduce(arr))
        Chem.SanitizeMol(mol, sanitize_param)
        return mol

    def _repr_svg_(self):
        """
        Modified to highlight bonds, from :
        https://github.com/rdkit/rdkit/blob/bc4d5478478542045a0f89e28a5da8b6d1f0e5d0/rdkit/Chem/Draw/IPythonConsole.py#L112
        """
        mol = self.rdkit
        return Chem.Draw._moltoSVG(
            mol,
            IPythonConsole.molSize,
            getattr(mol, "__sssAtoms", []),
            "",
            IPythonConsole.kekulizeStructures,
            drawOptions=IPythonConsole.drawOptions,
            highlightBonds=getattr(mol, "__sssBonds", []),
        )

    def get_substructure_match(self, smarts: str) -> Tuple[tuple]:
        """
        Apply RDKit `GetSubstructMatches` method.

        !!! note
            It highlight bonds as it seems to not working with the python API of RDKit.

        Args:
            smarts: the substructure to find.

        Returns:
            Tuple of atom ids for each match
        """
        substructure = Chem.MolFromSmarts(smarts)
        mol = self.rdkit
        matches = mol.GetSubstructMatches(substructure)
        highlight_atoms = getattr(mol, "__sssAtoms", [])
        highlight_bonds = [
            bond.GetIdx()
            for bond in mol.GetBonds()
            if self._is_highlight_bond(bond, highlight_atoms)
        ]
        setattr(self._rdkit, "__sssAtoms", highlight_atoms)
        setattr(self._rdkit, "__sssBonds", highlight_bonds)
        return matches

    def _is_highlight_bond(self, bond, highlight_atoms):
        return set(self._get_bond_atoms(bond)).issubset(set(highlight_atoms))

    @staticmethod
    def _get_bond_atoms(bond):
        return [bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()]

    def highlight_aromatics(self):
        """Highlight aromatics rings"""
        mol = self.rdkit
        ring_info = mol.GetRingInfo()
        aromatic_bonds = []
        aromatic_atoms = []
        for ring_bonds in ring_info.BondRings():
            for bond_idx in ring_bonds:
                bond = mol.GetBondWithIdx(bond_idx)
                if mol.GetBondWithIdx(bond_idx).GetIsAromatic():
                    aromatic_bonds.append(bond_idx)
                    aromatic_atoms += self._get_bond_atoms(bond)
        setattr(self._rdkit, "__sssAtoms", aromatic_atoms)
        setattr(self._rdkit, "__sssBonds", aromatic_bonds)
        return self

    def get_atoms(self):
        """Yield every atom in the molecule"""
        return self.rdkit.GetAtoms()

    def get_ring_systems(self, includeSpiro: bool = False) -> List[set]:
        """
        Get all ring systems.

        !!! note
            From [RDKit Cookbook]\
(https://www.rdkit.org/docs/Cookbook.html#count-ring-systems)

        Returns:
            List of set if atom id for each ring systems
        """
        return self._get_ring_systems(self.rdkit, includeSpiro=includeSpiro)

    @staticmethod
    def _get_ring_systems(mol, includeSpiro: bool = False):
        ring_info = mol.GetRingInfo()
        systems: List[Set] = []
        for ring in ring_info.AtomRings():
            ringAts = set(ring)
            nSystems = []
            for system in systems:
                nInCommon = len(ringAts.intersection(system))
                if nInCommon and (includeSpiro or nInCommon > 1):
                    ringAts = ringAts.union(system)
                else:
                    nSystems.append(system)
            nSystems.append(ringAts)
            systems = nSystems
        return systems



from __future__ import annotations
import re
from typing import List
from rdkit import Chem
from rdkit.Chem import AllChem
from chem_kit.molecule import Molecule
from ..chem_doodle import ChemDoodle
from .generator import SmartsFromSmiles, SmartsFromMetWorkV0
from .simplifier import TransformationSimplifier


class Transformation:
    """
    Instance of (bio-)transformation, using RDKit `Reaction` class.

    It's instantiate by providing SMARTS
    but it's also possible to `from_smiles()`
    and `from_metwork_v0()` class method.

    !!! example
        ```python
        from chem_kit import Transformation
        tsf = Transformation("[#6:1]-[#8:2]-[#1:3]\
>>[#6:1]-[#8:2]-[#6:3](-[#1])(-[#1])-[#1]")
        ```
    """

    def __init__(self, smarts: str):
        """
        Args:
            smarts: SMARTS of the transformation to create
        """
        self._set_from_smarts(smarts)

    def _set_from_smarts(self, smarts):
        self._smarts = smarts
        self._rdkit = AllChem.ReactionFromSmarts(smarts)

    @classmethod
    def from_smiles(cls, reactant: str, product: str) -> Transformation:
        """
        Args:
            reactant: SMILES of reactant
            product: SMILES of product

        Returns:
            Transformation instance
        """
        return cls(SmartsFromSmiles(reactant, product).smarts)

    @classmethod
    def from_metwork_v0(cls, smarts: str) -> Transformation:
        """
        Args:
            smarts: smarts from MetWork v0 reaction database

        Returns:
            Transformation instance
        """
        return cls(SmartsFromMetWorkV0(smarts).smarts)

    @property
    def smarts(self) -> str:
        """The SMARTS of the transformation."""
        return self._smarts

    @property
    def rdkit(self) -> Chem.rdChemReactions.ChemicalReaction:
        """
        The `Chem.rdChemReactions.ChemicalReaction` instance
        associate with the transformation.
        """
        return self._rdkit

    @property
    def reactant(self):
        return Molecule.from_rdkit(self.rdkit.GetReactants()[0])

    @property
    def product(self):
        return Molecule.from_rdkit(self.rdkit.GetProducts()[0])

    @property
    def mass_delta(self):
        return self.product.mass - self.reactant.mass

    @property
    def chemdoodle_json(self) -> dict:
        """
        [ChemDoodle JSON](https://web.chemdoodle.com/docs/chemdoodle-json-format)
        format of the transformation.
        """
        return ChemDoodle.react_to_json(self.rdkit)

    def _repr_png_(self):
        return self.rdkit._repr_png_()

    def _repr_svg_(self):
        return self.rdkit._repr_svg_()

    def run(self, smiles: str):
        """
        Run the transformation for a reactant.

        Args:
            smiles: SMILES of the reactant
        """
        react = Molecule(smiles).rdkit
        if re.search(r"(?:#1|H)", self.smarts):
            react = Chem.AddHs(react)
        products = self.rdkit.RunReactant(react, 0)
        res = []
        for r in products:
            try:
                res.append(Molecule.from_rdkit(r[0]))
            except Exception:
                pass
        return res

    def simplify(self, **params) -> List[Transformation]:
        """
        Simplify the transformation by removing non changing atoms that are not
        'connected' to the reaction site (i.e. atoms that changes).

        Args:
            params: See [SimplifierParams](transformation.md#simplifierparams)

        Returns:
            List of simplified transformations.
        """
        results = TransformationSimplifier(self.smarts, **params)
        return [
            self.__class__(smarts)._reindex_mapping()
            for smarts in results.simplified_smarts
        ]

    def reverse(self) -> Transformation:
        """
        Reverse the way of the transformation,
        i.e. reactant become product and product become reactant.
        """
        smarts = self.smarts.split(">>")
        smarts.reverse()
        self._set_from_smarts(">>".join(smarts))
        return self

    def _reindex_mapping(self):
        mapping_origin = {
            atom.GetAtomMapNum()
            for react in self.rdkit.GetReactants()
            for atom in react.GetAtoms()
        } - {0}
        mapping_origin = list(mapping_origin)
        mapping_origin.sort()
        mapping_change = {origin: new + 1 for new, origin in enumerate(mapping_origin)}
        smarts = ">>".join(
            [
                ".".join(
                    [
                        Chem.MolToSmarts(self._reindex_mol(mol, mapping_change))
                        for mol in mols
                    ]
                )
                for mols in [self.rdkit.GetReactants(), self.rdkit.GetProducts()]
            ]
        )
        self._set_from_smarts(smarts)
        return self

    @staticmethod
    def _reindex_mol(mol, mapping_change):
        for atom in mol.GetAtoms():
            map_num = atom.GetAtomMapNum()
            if map_num:
                atom.SetAtomMapNum(mapping_change.get(map_num, 0))
        return mol



import pkg_resources
from chem_kit.molecule import Molecule
from chem_kit.transformation import Transformation


__all__ = ["Molecule", "Transformation"]

__version__ = pkg_resources.get_distribution("chem_kit").version



from rdkit import Chem
from rdkit.Chem import rdFMCS
from chem_kit.molecule import Molecule, AROMATICITY_MODEL


class SmartsGenerator:

    EXPLICIT_H_MAP_FROM = 1000

    def __init__(self, *arg, **kwargs):
        self.smarts: str = self.get_smarts(*arg, **kwargs)

    def get_smarts(self, *source):
        mols = self.list_mols(*source)
        mols = [Chem.AddHs(mol) for mol in mols]
        matches = self.get_common_atoms(*mols)
        self.map_matches(matches, mols)
        h_mapped = self.get_h_mapped(mols)
        mols = [self.remove_h_non_mapped(mol) for mol in mols]
        mols = self.explicit_h_when_mapped_with_h(mols, h_mapped)
        mols_smarts = [Chem.MolToSmarts(mol) for mol in mols]
        return ">>".join(mols_smarts)

    @staticmethod
    def get_h_mapped(mols):
        h_mapped = []
        for mol in mols:
            for atom in mol.GetAtoms():
                map_num = atom.GetAtomMapNum()
                if atom.GetAtomicNum() == 1 and map_num != 0:
                    h_mapped.append(map_num)
        return h_mapped

    @staticmethod
    def explicit_h_when_mapped_with_h(mols, h_mapped):
        mols_res = []
        for mol in mols:
            mol = Chem.MolFromSmarts(Chem.MolToSmarts(mol))
            mol.UpdatePropertyCache(strict=False)
            idxs = [
                atom.GetIdx()
                for atom in mol.GetAtoms()
                if (atom.GetAtomMapNum() in h_mapped)
            ]
            if idxs:
                mol = Chem.AddHs(mol, onlyOnAtoms=idxs)
            mols_res.append(mol)
        return mols_res

    @staticmethod
    def list_mols(*smarts):
        mols = [Chem.MolFromSmarts(sm) for sm in smarts]
        for mol in mols:
            Chem.SanitizeMol(mol)
        return mols

    def get_common_atoms(self, *mols):
        mcs = rdFMCS.FindMCS(
            mols,
            atomCompare=rdFMCS.AtomCompare.CompareAny,
            # atomCompare=rdFMCS.AtomCompare.CompareElements,
            bondCompare=rdFMCS.BondCompare.CompareAny,
            # ringCompare=Chem.rdFMCS.RingCompare.StrictRingFusion,
            # ringCompare=Chem.rdFMCS.RingCompare.PermissiveRingFusion,
            # ringMatchesRingOnly=True,
            # completeRingsOnly=True,
        ).queryMol
        return [mol.GetSubstructMatch(mcs) for mol in mols]

    @staticmethod
    def map_matches(matches, mols):
        map_count = 1
        for atoms_idx in zip(*matches):
            atoms = [mol.GetAtomWithIdx(idx) for mol, idx in zip(mols, atoms_idx)]
            if set(atom.GetSymbol() for atom in atoms) != {"H"}:
                [atom.SetAtomMapNum(map_count) for atom in atoms]
                map_count += 1

    def remove_h_non_mapped(self, mol):
        rwmol = Chem.RWMol(mol)
        atoms_to_remove = []
        for atom in rwmol.GetAtoms():
            if atom.GetAtomMapNum() == 0 and atom.GetSymbol() == "H":
                atoms_to_remove.append(atom.GetIdx())
            if atom.GetAtomMapNum() > self.EXPLICIT_H_MAP_FROM:
                atom.SetAtomMapNum(0)
        atoms_to_remove.sort(reverse=True)
        for idx in atoms_to_remove:
            rwmol.RemoveAtom(idx)
        return rwmol


class SmartsFromSmiles(SmartsGenerator):
    def list_mols(self, *smiles):
        mols = [Molecule(sm, preserve_H=True).rdkit for sm in smiles]
        for mol in mols:
            h_map_idx = self.EXPLICIT_H_MAP_FROM + 1
            for atom in mol.GetAtoms():
                if atom.GetAtomicNum() == 1:
                    atom.SetAtomMapNum(h_map_idx)
                    h_map_idx += 1
        return mols


class SmartsFromMetWorkV0(SmartsGenerator):
    def get_smarts(self, smarts):
        smarts = smarts.replace("=,:", "=")
        smarts = smarts.split(">>")
        mols = [Chem.MolFromSmarts(sm) for sm in smarts]
        mols = [self.unset_mapping(mol) for mol in mols]
        for mol in mols:
            Chem.Kekulize(mol)
            Chem.SetAromaticity(mol, AROMATICITY_MODEL)
        return super().get_smarts(*mols)

    @staticmethod
    def list_mols(*source):
        return source

    @staticmethod
    def unset_mapping(mol):
        for atom in mol.GetAtoms():
            atom.SetAtomMapNum(0)
        return mol



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chem_kit', 'chem_kit.transformation']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.2.0,<9.0.0',
 'ipython>=7.21.0,<8.0.0',
 'numpy>=1.19.5,<2.0.0',
 'pandas>=1.2.1,<2.0.0',
 'pydantic>=1.8.1,<2.0.0']

setup_kwargs = {
    'name': 'chem-kit',
    'version': '0.1.4',
    'description': 'A chemical toolbox based on RDKit',
    'long_description': '# ChemKit\n\n<p align="center">\n<a href="https://pypi.org/project/chem-kit" target="_blank">\n    <img src="https://img.shields.io/pypi/v/chem-kit?color=%2334D058&label=pypi%20package" alt="Package version">\n</a>\n</p>\n\n---\n\n**Documentation**: <a href="http://chem-kit.metwork.science/" target="_blank">http://chem-kit.metwork.science/</a>\n\n**Source Code**: <a href="https://github.com/metwork-project/chem-kit" target="_blank">https://github.com/metwork-project/chem-kit</a>\n\n---\n\nChemKit is a chemical toolbox based on [RDKit](https://www.rdkit.org/) with currently 2 main purposes :\n\n- Facilitate the usage of the [RDKIt Python API](https://www.rdkit.org/docs/api-docs.html)\n with some more easy to use classes that can occasionally fix some bug (especially with Jupyter rendering).\n\n- Provide tailored methods for the [MetWork](http://www.metwork.science) project\n\n## Usage\n\n### Manipulate Molecules\n\n```python\n    from chem_kit import Molecule\n    mol = Molecule("CCO")\n```\n\n### Manipulate Transformation\n\n```python\n    from chem_kit import Transformation\n    tsf = Transformation("[#6:1]-[#8:2]-[#1:3]>>[#6:1]-[#8:2]-[#6:3](-[#1])(-[#1])-[#1]")\n```\n\nMore examples with [Jupyter notebook](http://chem-kit.metwork.science/jupyter_example/)\n\n## Install\n\nLike RDKit, ChemKit needs [Conda](https://docs.conda.io) :\n\n```bash\nconda env create -f conda-env.yml\nconda activate chem_kit\n```\n\nTo manage other required Python packages, \nthe better way is to use [Poetry](https://python-poetry.org) on top of Conda :\n\n```bash\ncd /path/to/chem_kit\npoetry install\n```\n\n> Poetry manipulate Python packages directly on Conda env.\n',
    'author': 'Yann Beauxis',
    'author_email': 'dev@yannbeauxis.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/metwork-project/chem-kit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



import re
from typing import List, Set
from pydantic import BaseModel, Field
from rdkit import Chem
from rdkit.Chem import rdFMCS
from chem_kit.molecule import Molecule


class SimplifierParams(BaseModel):
    hetero_atoms: bool = Field(
        True,
        description="Include hetero atoms close to transformation site",
    )
    aromatic: bool = Field(
        True,
        description="Include aromatic cycles close to transformation site",
    )
    conjugated: bool = Field(
        False,
        description="Include conjugated bonds close to transformation site",
    )
    cycles: bool = Field(
        False,
        description="Include cycles close to transformation site",
    )


class TransformationSimplifier:

    simplified_smarts: List[str] = []

    def __init__(self, smarts, **params):
        self.params = SimplifierParams(**params)
        self._full_smarts = smarts
        self.simplify_smarts()

    @property
    def full_smarts(self):
        return self._full_smarts

    def simplify_smarts(self):
        mols = [Chem.MolFromSmarts(sm) for sm in self.full_smarts.split(">>")]
        mcs = rdFMCS.FindMCS(mols).queryMol

        _ = [mol.GetSubstructMatch(mcs) for mol in mols]

        mol_simplifiers = [MoleculeSimplifier(mol, self.params) for mol in mols]

        for ms in mol_simplifiers:
            ms.init_keep()

        self.keep_for_bond_diff(mol_simplifiers)

        for ms in mol_simplifiers:
            ms.propagate_map_to_keep()

        self.keep_for_bond_diff(mol_simplifiers)

        self.join_mapped_atoms(mol_simplifiers)

        smarts = [Chem.MolToSmarts(ms.mol) for ms in mol_simplifiers]
        splits = [sm.split(".") for sm in smarts]

        splits = [
            [
                {
                    "smarts": sm,
                    "mapped": set(re.findall(r"\[#\d+(?:\&(?:\++|-))?:(\d+)\]", sm)),
                }
                for sm in split
            ]
            for split in splits
        ]
        smarts = []
        for dic1 in splits[0]:
            for dic2 in splits[1]:
                if dic1["mapped"] == dic2["mapped"]:
                    smarts.append(">>".join([dic1["smarts"], dic2["smarts"]]))

        smarts = list(set(smarts) - {">>"})

        self.simplified_smarts = smarts

    def join_mapped_atoms(self, mol_simplifiers):
        map_to_keep = set([])
        for ms in mol_simplifiers:
            map_to_keep = map_to_keep | ms.map_to_keep
        for ms in mol_simplifiers:
            ms.map_to_keep = map_to_keep
            ms.remove_atoms()

    @staticmethod
    def keep_for_bond_diff(mol_simplifiers):
        map_to_keep = set()
        bond_types = [ms.get_bond_types() for ms in mol_simplifiers]
        bond_types_keys = [set(bond_type.keys()) for bond_type in bond_types]
        bond_types_keys = set().union(*bond_types_keys)
        for key in bond_types_keys:
            values = set([bond_type.get(key, tuple()) for bond_type in bond_types])
            if len(values) > 1:
                map_to_keep = map_to_keep | set(key)
        for ms in mol_simplifiers:
            ms.map_to_keep = ms.map_to_keep | map_to_keep


class MoleculeSimplifier:
    def __init__(self, mol, params: SimplifierParams):
        self.params = params
        mol = Molecule.resolve_charge(mol)
        self.mol = mol
        self.mol_smiles = Molecule(Chem.MolToSmiles(mol), preserve_H=True)
        self.map_to_keep: Set = set()

    def append_map_to_keep(self, value):
        self.map_to_keep = self.map_to_keep | {value}

    def init_keep(self):
        matched_atoms = getattr(self.mol, "__sssAtoms")
        all_idx = {atom.GetIdx() for atom in self.mol.GetAtoms()}
        idx_to_keep = all_idx - set(matched_atoms)
        self.map_to_keep = {
            self.mol.GetAtomWithIdx(idx).GetAtomMapNum() for idx in idx_to_keep
        } | {0}

    def propagate_map_to_keep(self):
        self.propagate_hetero(to_connector=True)
        if self.params.cycles:
            self.propagate_ring()
            self.propagate_hetero(to_connector=False)

    def propagate_hetero(self, to_connector):
        mol_smiles = self.mol_smiles.rdkit
        for idx in self.idx_to_keep_from_map(mol_smiles):
            atom = mol_smiles.GetAtomWithIdx(idx)
            visitor = AtomVisitor(self, atom, params=self.params)
            visitor.propagate(to_connector=to_connector)

    def propagate_ring(self):
        mol_smiles = self.mol_smiles
        systems = mol_smiles.get_ring_systems()
        idx_to_keep = set([])

        for atom in mol_smiles.get_atoms():
            if atom.GetAtomMapNum() in self.map_to_keep:
                for system in systems:
                    if atom.GetIdx() in system:
                        idx_to_keep = idx_to_keep | system

        map_to_keep = self.map_to_keep
        systems_to_keep = {
            mol_smiles.rdkit.GetAtomWithIdx(idx).GetAtomMapNum() for idx in idx_to_keep
        }
        map_to_keep = map_to_keep | systems_to_keep
        self.map_to_keep = map_to_keep

    def get_idx_to_remove(self):
        mol = self.mol
        idx_to_keep = self.idx_to_keep_from_map(mol)
        all_idx = {atom.GetIdx() for atom in mol.GetAtoms()}
        idx_to_remove = list(all_idx - idx_to_keep)
        idx_to_remove.sort(reverse=True)
        return idx_to_remove

    def remove_atoms(self):
        idx_to_remove = self.get_idx_to_remove()
        rwmol = Chem.RWMol(self.mol)
        for idx in idx_to_remove:
            rwmol.RemoveAtom(idx)
        self.mol = rwmol

    def idx_to_keep_from_map(self, mol):
        return {
            atom.GetIdx()
            for atom in mol.GetAtoms()
            if atom.GetAtomMapNum() in self.map_to_keep
        }

    def get_bond_types(self):
        bond_types = dict()
        for bond in self.mol.GetBonds():
            mapped_atoms = [
                bond.GetBeginAtom().GetAtomMapNum(),
                bond.GetEndAtom().GetAtomMapNum(),
            ]
            mapped_atoms.sort()

            if all(mapped_atoms):
                bond_types[tuple(mapped_atoms)] = bond.GetBondType()
        return bond_types


class AtomVisitor:
    def __init__(self, mol_simplifier, atom, params: SimplifierParams):
        self.params = params
        self.mol_simplifier = mol_simplifier
        self.atom = atom

    @property
    def is_hetero(self):
        return self.atom.GetAtomicNum() != 6

    @property
    def is_aromatic(self):
        return self.atom.GetIsAromatic()

    @property
    def map_num(self):
        return self.atom.GetAtomMapNum()

    @property
    def is_keeped(self):
        return self.map_num in self.mol_simplifier.map_to_keep

    def to_propagate_from(self, from_visitor, bond, to_connector):
        if from_visitor.is_keeped and self.is_keeped:
            return False, False
        if from_visitor.is_aromatic and self.is_aromatic and self.params.aromatic:
            return True, False
        if bond.GetIsConjugated() and self.params.conjugated:
            return True, to_connector
        if self.is_hetero and self.params.hetero_atoms:
            return True, to_connector
        if from_visitor.is_hetero and self.params.hetero_atoms and to_connector:
            return True, self.is_hetero
        return to_connector, False

    def set_keep(self):
        self.mol_simplifier.append_map_to_keep(self.map_num)

    def propagate(self, to_connector):
        for atom, bond in self.connected_atoms():
            visitor = self.__class__(self.mol_simplifier, atom, params=self.params)
            propagate, to_connector = visitor.to_propagate_from(
                self, bond, to_connector=to_connector
            )
            if propagate:
                visitor.set_keep()
                visitor.propagate(to_connector=to_connector)

    def connected_atoms(self):
        for bond in self.atom.GetBonds():
            for atom in (bond.GetBeginAtom(), bond.GetEndAtom()):
                if atom.GetIdx() != self.atom.GetIdx():
                    yield atom, bond
